// ListView.cp

#ifndef ListView_h
#include "ListView.h"
#endif
#ifndef ViewMap_h
#include "ViewMap.h"
#endif
#ifndef ViewCell_h
#include "ViewCell.h"
#endif

ListView::ListView()
  : cellHeight( 0 ),
	 cellCount( 0 )
  {
  }

Rectangle ListView::CellArea( const ViewMap& map, uint32 cell ) const
  {
	Assert( cellHeight > 0 );
	Assert( cell < cellCount );

	return Rectangle( map.Bounds().left,
							map.Bounds().top + cell * cellHeight,
							map.Bounds().right,
							map.Bounds().top + cell * cellHeight + cellHeight );
  }

Rectangle ListView::CellArea( const ViewMap& map, Range<uint32> cells ) const
  {
	Assert( cellHeight > 0 );
	Assert( cells.End() <= cellCount );

	return Rectangle( map.Bounds().left,
							map.Bounds().top + cells.Start() * cellHeight,
							map.Bounds().right,
							map.Bounds().top + cells.End() * cellHeight );
  }

Rectangle ListView::UnusedArea( const ViewMap& map ) const
  {
	return Rectangle( map.Bounds().left,
							map.Bounds().top + cellCount * cellHeight,
							map.Bounds().right,
							map.Bounds().bottom );
  }

void ListView::Draw( const ViewMap& toDraw ) const
  {
	Assert( cellHeight > 0 );
	
	uint32 index = PixelToCell( toDraw.VisibleBounds().top - toDraw.Bounds().top );
	
	Rectangle area( CellArea( toDraw, index ) );

	  {
		ViewMap cellMap( toDraw );
	
		while ( index < cellCount && area.top < toDraw.VisibleBounds().bottom )
		  {
			cellMap.Set( toDraw, area );
			(*this)[index].Draw( cellMap );
			
			index++;
			area += PointObject( 0, cellHeight );
		  }
	  }
	
	area.bottom = toDraw.Bounds().bottom;
	toDraw.Erase( area );
  }

void ListView::SetCellHeight( uint32 height )
  {
	cellHeight = height;
	Assert( CanMultiply( cellHeight, cellCount ) );
	View::Invalidate();
  }

void ListView::AddedCells( Range<uint32> range )
  {
	Assert( CanAdd( range.Length(), cellCount ) );
	cellCount += range.Length();
	Assert( CanMultiply( cellHeight, cellCount ) );
	Invalidate( Range<uint32>( range.Start(), cellCount ) );
  }

void ListView::RemovedCells( Range<uint32> range )
  {
	Assert( range.End() <= cellCount );
	Invalidate( Range<uint32>( range.Start(), cellCount ) );
	cellCount -= range.Length();
  }

void ListView::Update( uint32 cell ) const
  {
	if ( !Mapped() )
		return;

	ViewMap map( *this );
	View::Update( CellArea( map, cell ) );
  }

void ListView::Redraw( uint32 cell ) const
  {
	if ( !Mapped() )
		return;

	ViewMap map( *this );
	View::Redraw( CellArea( map, cell ) );
  }

void ListView::Invalidate( uint32 cell ) const
  {
	if ( !Mapped() )
		return;

	ViewMap map( *this );
	map.Invalidate( CellArea( map, cell ) );
  }

void ListView::Update( Range<uint32> range ) const
  {
	if ( !Mapped() )
		return;

	ViewMap map( *this );
	View::Update( CellArea( map, range ) );
  }

void ListView::Redraw( Range<uint32> range ) const
  {
	if ( !Mapped() )
		return;

	ViewMap map( *this );
	View::Redraw( CellArea( map, range ) );
  }

void ListView::Invalidate( Range<uint32> range ) const
  {
	if ( !Mapped() )
		return;

	ViewMap map( *this );
	map.Invalidate( CellArea( map, range ) );
  }
